/*
    The contents of this file were written by <bofn@irq.org>
    (except where noted) and are covered by his licence
    (see bofn.licence.txt in this distribution.

    I'd like to thank him for allowing me to include these
    functions in dellfand. They are the ones that do the
    'magic', and I havn't made any effort to discover how
    they work.
*/


#define D_GET_BATT      0x003E
#define D_FN_STATUS		0x0025
#define D_POWER_STATUS	0x0069
#define D_SET_FAN			0x01a3
#define D_GET_FAN			0x00a3
#define D_GET_SPEED		0x02a3
#define D_GET_TEMP		0x10a3

#define FAN_MULT		30
#define MAX_TEMP		127

#define INIT_REGS   DRegs regs = { 0, 0, 0, 0, 0, 0 }

//-------------------------

typedef struct {
    unsigned int eax;
    unsigned int ebx __attribute__ ((packed));
    unsigned int ecx __attribute__ ((packed));
    unsigned int edx __attribute__ ((packed));
    unsigned int esi __attribute__ ((packed));
    unsigned int edi __attribute__ ((packed));
} DRegs;

//-------------------------

static int d_io(DRegs *regs)
{
    int rc;
    int eax = regs->eax;

// LW 20061210: alternative code for 64 bit, not part
// of original file, thanks to 
// Facundo <adscvr@gmail.com> for pointing me
// at this code.

#if defined (__x86_64__)
        asm("pushq %%rax\n\t"
          "movl 0(%%rax),%%edx\n\t"
          "pushq %%rdx\n\t"
          "movl 4(%%rax),%%ebx\n\t"
          "movl 8(%%rax),%%ecx\n\t"
          "movl 12(%%rax),%%edx\n\t"
          "movl 16(%%rax),%%esi\n\t"
          "movl 20(%%rax),%%edi\n\t"
          "popq %%rax\n\t"
          "out %%al,$0xb2\n\t"
          "out %%al,$0x84\n\t"
          "xchgq %%rax,(%%rsp)\n\t"
          "movl %%ebx,4(%%rax)\n\t"
          "movl %%ecx,8(%%rax)\n\t"
          "movl %%edx,12(%%rax)\n\t"
          "movl %%esi,16(%%rax)\n\t"
          "movl %%edi,20(%%rax)\n\t"
          "popq %%rdx\n\t"
          "movl %%edx,0(%%rax)\n\t"
          "lahf\n\t"
          "shrl $8,%%eax\n\t"
          "andl $1,%%eax\n"
	  :"=a"(rc)
          :    "a"(regs)
          :    "%ebx", "%ecx", "%edx", "%esi", "%edi", "memory");
#else
 asm("pushl %%eax\n\t" \
      "movl 0(%%eax),%%edx\n\t" \
	   "push %%edx\n\t" \
	   "movl 4(%%eax),%%ebx\n\t" \
	   "movl 8(%%eax),%%ecx\n\t" \
	   "movl 12(%%eax),%%edx\n\t" \
	   "movl 16(%%eax),%%esi\n\t" \
	   "movl 20(%%eax),%%edi\n\t" \
	   "popl %%eax\n\t" \
	   "out %%al,$0xb2\n\t" \
	   "out %%al,$0x84\n\t" \
	   "xchgl %%eax,(%%esp)\n\t"
	   "movl %%ebx,4(%%eax)\n\t" \
	   "movl %%ecx,8(%%eax)\n\t" \
	   "movl %%edx,12(%%eax)\n\t" \
	   "movl %%esi,16(%%eax)\n\t" \
	   "movl %%edi,20(%%eax)\n\t" \
	   "popl %%edx\n\t" \
	   "movl %%edx,0(%%eax)\n\t" \
	   "lahf\n\t" \
	   "shrl $8,%%eax\n\t" \
	   "andl $1,%%eax\n" \
	    : "=a" (rc)
	    : "a" (regs)
	    : "%ebx", "%ecx", "%edx", "%esi", "%edi", "memory");
#endif

    // LW 20061222
    // Casted to unsigned int to remove compiler warning,
    // not part of original code

    if (
		   (rc != 0) || 
		   ((regs->eax & 0xffff) == 0xffff) || 
		   (regs->eax == (unsigned int)eax)) 
    {
    	return -1;
    }

 return 0;
}

//------------------------------------------------

static int get_fan_status(int fan)
{
    INIT_REGS;
    int rc;

    regs.eax = D_GET_FAN;
    regs.ebx = fan & 0xff;
    if ((rc=d_io(&regs)) < 0) 
    {
	   return rc;
    }

 return (regs.eax & 0xff);
}

//==

static int get_fan_speed(int fan)
{
    INIT_REGS;
    int rc;

    regs.eax = D_GET_SPEED;
    regs.ebx = fan & 0xff;
    if ((rc=d_io(&regs)) < 0) 
    {
  	   return rc;
    }

 return (regs.eax & 0xffff) * FAN_MULT;
}

//==

static int set_fan(int fan, int speed)
{
    INIT_REGS;
    int rc;

    speed = (speed < 0) ? 0 : ((speed > 2) ? 2 : speed);

    regs.eax = D_SET_FAN;
    regs.ebx = (fan & 0xff) | (speed << 8);
    if ((rc=d_io(&regs)) < 0) 
    {
	   return rc;
    }

 return (get_fan_status(fan));
}

static int get_cpu_temp(void)  // in 'Celcius'
{
    INIT_REGS;
    int rc;
    int temp;

    regs.eax = D_GET_TEMP;
    if ((rc=d_io(&regs)) < 0) 
    {
	   return rc;
    }
    
    temp = regs.eax & 0xff;

 return temp;
}
